#!/usr/bin/perl -Tw
#####################################################
# gdipupdt.cgi
#
# This is the GnuDIP HTTP update server.
#
# See COPYING for licensing information.
#
# Written by:
#
#   Creighton MacDonnell
#   http://macdonnell.ca/
#
#####################################################

# Perl modules
use strict;
use CGI;

# global variables
use vars qw($conf $gnudipdir);
use vars qw($pref $remote_ip);
use vars qw($reqparm $thiscgi $logger);

# locate ourselves
use FindBin;
BEGIN {
  $gnudipdir = '';
  if ($FindBin::Bin =~ /(.*)\/.+?/) {
    $gnudipdir = $1;
  }
}
use lib "$gnudipdir/lib";

# GnuDIP common subroutines
use gdiplib;
use dbprefs;
use dbusers;

gdipupdt();

### below this will later become a module? ###

# variables global to this file
my ($pretty_query, $logprfx);

sub gdipupdt {

  # get preferences from config file
  $conf = getconf();
  if (! $conf) {
    print STDERR "GnuDIP IP Update CGI has exited - getconf returned nothing\n";
    dberror();
  }

  # logger command
  $logprfx = '';
  $logger = $$conf{'logger_updt'};
  if (! defined $logger) {
    $logprfx = 'HTTP Update: ';
    $logger = $$conf{'logger_cgi'};
  }

  # not called by HTTP server
  if (!$ENV{'REQUEST_METHOD'} or
      $ENV{'REQUEST_METHOD'} ne 'GET' and
      $ENV{'REQUEST_METHOD'} ne 'POST') {
    writelog( 'gdipupdt.cgi: REQUEST_METHOD environment variable is not set');
    print "REQUEST_METHOD environment variable is not set\n";
    writelog( 'gdipupdt.cgi: Not called by HTTP server');
    print "Not called by HTTP server\n";
    exit;
  }

  # IP address of connecting machine
  $remote_ip   = $ENV{'REMOTE_ADDR'};
  $remote_ip = '0.0.0.0' if !$remote_ip;

  # CGI request information
  my $query = new CGI;
  $thiscgi  = $query->url();
  my %reqparmhash;
  $reqparm  = \%reqparmhash;
  foreach my $name ($query->param()) {
    $$reqparm{$name} = join("\0", $query->param($name));
  }

  # log the HTTP request info to HTTP server log 
  # - for debugging
  #logreq();

  # get preferences from the database
  $pref = getprefs();

  # called as GET with no parms?
  gen_salt()
    if $ENV{'REQUEST_METHOD'} eq 'GET' and !$ENV{'QUERY_STRING'};

  # html for query string
  $pretty_query = qq*
<p>
<font size="+1"><b>Query String</b></font>
<p>
$ENV{'QUERY_STRING'}
*;

  # called as GET with all all parms?
  do_updt()
    if $ENV{'REQUEST_METHOD'} eq 'GET' and
       defined $$reqparm{'salt'} and
       defined $$reqparm{'time'} and
       defined $$reqparm{'sign'} and
       defined $$reqparm{'domn'} and
       defined $$reqparm{'user'} and
       defined $$reqparm{'pass'} and
       defined $$reqparm{'reqc'};

  # bad HTTP request
  writelog($logprfx .
    "Invalid query string from $remote_ip");
  pg_updtsrv(tst(qq*
Error: Invalid query string
$pretty_query
*));
}

# generate salt response and exit
sub gen_salt {

  my $salt     = randomsalt();
  my $gentime  = time();
  my $checkval = md5_hex("$salt.$gentime.$$pref{'SERVER_KEY'}");

  pg_updtsrv(
    'Salt generated',
    'salt', $salt,
    'time', $gentime,
    'sign', $checkval
    );
}

sub do_updt {

  my $salt         = $$reqparm{'salt'};
  my $gentime      = $$reqparm{'time'};
  my $ip           = $remote_ip;
  my $clientuser   = $$reqparm{'user'};
  my $clientpass   = $$reqparm{'pass'};
  my $clientdomain = $$reqparm{'domn'};
  my $clientaction = $$reqparm{'reqc'};
  my $clientip     = $$reqparm{'addr'};

  # validate the signature
  my $checkval = md5_hex("$salt.$gentime.$$pref{'SERVER_KEY'}");
  if ($checkval ne $$reqparm{'sign'}) {
    writelog($logprfx .
      "Invalid signature from $remote_ip: user $clientuser.$clientdomain");
    respond(tst(qq*
Error: Invalid signature
$pretty_query
*),
      '1');
  }

  # seconds to wait for response to prompt
  my $timeout = $$conf{'timeout_updt'};
  if (! defined $timeout) {
    $timeout = 6 * $$conf{'timeout'};
  }

  # timed out?
  if (time() > $gentime + $timeout) {
    writelog($logprfx .
      "Salt value too old from $remote_ip: user $clientuser.$clientdomain");
    respond('Error: Salt value too old', '1');
  }

  # sensible request?
  if($clientaction ne '0' && $clientaction ne '1' && $clientaction ne '2') {
    writelog($logprfx .
      "Invalid client request code from $ip: user $clientuser.$clientdomain");
    respond(tst(qq*
Error: Invalid client request code
$pretty_query
*),
      '1');
  }

  # retrieve user information
  my $userinfo = getuser($clientuser, $clientdomain);

  # bad login?
  if (!$userinfo                              or
      $clientuser   ne $$userinfo{'username'} or
      $clientdomain ne $$userinfo{'domain'}   or
      $clientpass   ne md5_hex("$$userinfo{'password'}.$salt")
      ) {
    writelog($logprfx .
      "Invalid login attempt from $ip: user $clientuser.$clientdomain");
    respond(tst(qq*
Error: Invalid login attempt
$pretty_query
*),
      '1');
  }

  # update request with 0.0.0.0 for address? convert to offline
  $clientaction = '1'
    if $clientaction eq '0' and
       defined($clientip)   and $clientip eq '0.0.0.0';

  # client passed an IP address?
  if (!defined($clientip) or $clientip eq '') {
    writelog($logprfx .
      "No IP address passed from $ip: user $clientuser.$clientdomain")
      if $clientaction eq '0';
    $clientip = $ip;
  }

  # invalid IP address?
  if($clientaction ne '1' && !validip($clientip)) {
    writelog($logprfx .
      "Unserviceable IP address from $ip: user $clientuser.$clientdomain - IP: $clientip");
    respond(tst(qq*
Error: Unserviceable IP address - IP: $clientip
$pretty_query
*),
      '1');
    exit;
  }

  # TTL value
  my $TTL = 0;
  $TTL = $$conf{'TTL'} if $$conf{'TTL'};
  $TTL = $$conf{"TTL.$clientdomain"}
      if $$conf{"TTL.$clientdomain"};

  # a modify request?
  if ($clientaction eq '0' or $clientaction eq '2') {

    # IP address unchanged?
    if ($$userinfo{'currentip'} eq $clientip) {
      writelog($logprfx .
        "User $clientuser.$clientdomain remains at ip $clientip");
      updateuser($userinfo);

    # do the update
    } else {
      donsupdate($clientdomain,
        "update delete $clientuser.$clientdomain. A",
        "update add    $clientuser.$clientdomain. $TTL A $clientip");
      writelog($logprfx .
        "User $clientuser.$clientdomain successful update to ip $clientip");
      $$userinfo{'currentip'} = $clientip;
      updateuser($userinfo);
    }

    if ($clientaction eq '2') {
      respond('Successful update request', '0', $clientip);
    } else {
      respond('Successful update request', '0');
    }

  # an offline request
  } else {

    # IP address unchanged?
    if ($$userinfo{'currentip'} eq '0.0.0.0') {
      writelog($logprfx .
        "User $clientuser.$clientdomain remains removed");
      updateuser($userinfo);

    # do the update
    } else {
      donsupdate($clientdomain,
        "update delete $clientuser.$clientdomain. A");
      writelog($logprfx .
        "User $clientuser.$clientdomain successful remove from ip $clientip");
      $$userinfo{'currentip'} = '0.0.0.0';
      updateuser($userinfo);
    }

    respond('Successful offline request', '2');
  }
}

#############################################################
# common subroutines
#############################################################

# print response page
sub respond {
  my $msg  = shift;
  my $retc = shift;
  my $addr = shift;
  pg_updtsrv($msg, 'retc', $retc, 'addr', $addr) if $addr;
  pg_updtsrv($msg, 'retc', $retc);
}

# call nsupdate
sub donsupdate {
  if (!callnsupdate(@_)) {
    writelog($logprfx .
      "gdipupdt.cgi has exited - callnsupdate failed");
    bad_config();
  }
}

# write to log
sub writelog {
  if (!calllogger($logger, @_)) {
    print STDERR "gdipupdt.cgi has exited - calllogger failed\n";
    bad_config();
  }
}

# called for database error
sub dberror {
  bad_config();
}

# configuration error message
sub bad_config {
  pg_updtsrv('Error: GnuDIP Configuration or Interface Problem');
}

# generate page with optional buried meta tags
sub pg_updtsrv {
  my $msg = shift;
  my @meta = @_;
  tpr(qq*
Content-Type: text/html; charset=us-ascii

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
                      "http://www.w3.org/TR/html4/loose.dtd\">
<html>
<head>
<title>
GnuDIP Update Server
</title>
*);
  while (my $name = shift @meta) {
    my $content = shift @meta;
    tpr(qq*
<meta name="$name" content="$content">
*);
  }
  tpr(qq*
</head>
<body>
<center>
<h2>
GnuDIP Update Server
</h2>
$msg
</center>
</body>
</html>
*);
  exit;  
}

# display the CGI data in the HTTP server log
sub logreq {
  my $var;
  my $val;
  print STDERR "ENV:\n";
  foreach $var (sort(keys(%ENV))) {
    $val = $ENV{$var};
    $val =~ s|\n|\\n|g;
    $val =~ s|"|\\"|g;
    print STDERR "  ${var}=\"${val}\"\n";
  }
  print STDERR "reqparm:\n";
  foreach $var (sort(keys(%$reqparm))) {
    $val = $$reqparm{$var};
    $val =~ s|\n|\\n|g;
    $val =~ s|"|\\"|g;
    print STDERR "  ${var}=\"${val}\"\n";
  }
}

